library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.STD_LOGIC_UNSIGNED.all;

 
entity cfide is
   port ( 
	cpuclk: in std_logic;	
	n_reset: in std_logic;	
	cpuena_in: in std_logic;			
	memdata_in: in std_logic_vector(15 downto 0);		
	addr: in std_logic_vector(23 downto 0);		
	cpudata_in: in std_logic_vector(15 downto 0);	
	state: in std_logic_vector(1 downto 0);		
	lds: in std_logic;			
	uds: in std_logic;			
	user_in: in std_logic_vector(7 downto 0);		
	idedata_in: in std_logic_vector(15 downto 0);		
	floppy_in: in std_logic_vector(15 downto 0);		
	statfloppy: in std_logic_vector(7 downto 0);		
	mdb: in std_logic_vector(7 downto 0);		
	sclk		: in std_logic;
	sd_di		: in std_logic;
		
	idedata: out std_logic_vector(15 downto 0);		
	idea: out std_logic_vector(2 downto 0);		
	ide_wr: out std_logic;			
	ide_rd: out std_logic;			
	ide_cs0: out std_logic;			
	ide_cs1: out std_logic;			
	ide_wrena: buffer std_logic;	
	ide_wrdata: out std_logic;
			
	memwr: out std_logic;			
	cpudata: out std_logic_vector(15 downto 0);		
	cpuena: buffer std_logic;			
	TxD: out std_logic;			
	sd_cs 		: out std_logic_vector(7 downto 0);
	sd_clk 		: out std_logic;
	sd_do		: out std_logic;
	user: out std_logic_vector(7 downto 0);		
	floppy: out std_logic_vector(15 downto 0)		
   );

end cfide;


architecture wire of cfide is

	COMPONENT startram
    PORT 
	(
		address		: IN STD_LOGIC_VECTOR (12 DOWNTO 0);
		byteena		: IN STD_LOGIC_VECTOR (1 DOWNTO 0);
		clock		: IN STD_LOGIC ;
		data		: IN STD_LOGIC_VECTOR (15 DOWNTO 0);
		wren		: IN STD_LOGIC ;
		q		: OUT STD_LOGIC_VECTOR (15 DOWNTO 0)
    );
	END COMPONENT;




signal partshift: std_logic_vector(39 downto 0);
signal mdbd: std_logic;
signal mdcdd: std_logic;
signal mdcd: std_logic;
signal shift: std_logic_vector(9 downto 0);
signal clkgen: std_logic_vector(9 downto 0);
--signal stop: std_logic;
signal shiftout: std_logic;
signal txbusy: std_logic;
signal ld: std_logic;
--signal io_select: std_logic;
signal rs232_select: std_logic;
signal KEY_select: std_logic;
signal USER_select: std_logic;
signal FLOPPY_select: std_logic;
signal PART_select: std_logic;
signal SPI_select: std_logic;
signal ROM_select: std_logic;
signal RAM_write: std_logic;
--signal rom_select: std_logic;
signal rs232data: std_logic_vector(15 downto 0);
signal part_in: std_logic_vector(15 downto 0);
signal IOdata: std_logic_vector(15 downto 0);
signal IDErd_data: std_logic_vector(15 downto 0);
signal IOcpuena: std_logic;
signal ide_select: std_logic;

type micro_states is (idle, nop, io_aktion, io_rd_data, ide1, ide2, ide3, ide4, ide5, ide6, ide7);
signal micro_state		: micro_states;
signal next_micro_state		: micro_states;

signal sd_out	: std_logic_vector(7 downto 0);
signal sd_di_in	: std_logic;
signal shiftcnt	: std_logic_vector(3 downto 0);
signal sck		: std_logic;
signal zscs		: std_logic;
--signal zexe_dir		: std_logic;
signal scs		: std_logic_vector(7 downto 0);
--signal SDcommand	: std_logic_vector(7 downto 0);
signal SD_busy		: std_logic;
--signal links_tmp	: std_logic_vector(15 downto 0);
--signal rechts_tmp	: std_logic_vector(7 downto 0);
--signal clkgen: std_logic_vector(10 downto 0);
signal spi_div: std_logic_vector(7 downto 0);
signal spi_speed: std_logic_vector(7 downto 0);
signal rom_data: std_logic_vector(15 downto 0);



begin

srom: startram
	PORT MAP 
	(
--		address => addr(11 downto 4)&'0'&addr(3 downto 1),	--: IN STD_LOGIC_VECTOR (11 DOWNTO 0);
		address => addr(13 downto 1),	--: IN STD_LOGIC_VECTOR (11 DOWNTO 0);
		byteena(0)	=> not lds,			--	: IN STD_LOGIC_VECTOR (1 DOWNTO 0),
		byteena(1)	=> not uds,			--	: IN STD_LOGIC_VECTOR (1 DOWNTO 0),
		clock   => not cpuclk,								--: IN STD_LOGIC ;
		data	=> cpudata_in,		--	: IN STD_LOGIC_VECTOR (15 DOWNTO 0),
		wren	=> RAM_write,		-- 	: IN STD_LOGIC ,
		q		=> rom_data									--: OUT STD_LOGIC_VECTOR (15 DOWNTO 0)
    );




memwr <= '0' WHEN ROM_select='0' AND addr(23)='0' ELSE '1';
cpudata <=  rom_data WHEN ROM_select='1' ELSE 
			IOdata WHEN IOcpuena='1' ELSE
			statfloppy&user_in WHEN USER_select='1' ELSE 
			floppy_in WHEN FLOPPY_select='1' ELSE 
			part_in WHEN PART_select='1' ELSE 
			memdata_in;
part_in <=  partshift(39 downto 24) WHEN addr(2 downto 1)="00" ELSE			
			partshift(23 downto 8) WHEN addr(2 downto 1)="01" ELSE			
			partshift(7 downto 0)&"XXXXXXXX"; 
--cpudata <= romdata WHEN rom_select='1' ELSE memdata_in WHEN addr(23)='0' ELSE IOdata;
--cpudata <= memdata_in WHEN addr(23)='0' ELSE IOdata;
IOdata <=   --rs232data WHEN rs232_select='1' ELSE 
			SD_busy&"0000000"&SD_out WHEN SPI_select='1' ELSE
			IDErd_data(7 downto 0)&IDErd_data(15 downto 8);
--cpuena <= '1' WHEN (cpuena_in='1' AND (addr(23)='0' OR state="01")) OR micro_state=io_aktion ELSE '0'; 
--cpuena <= '1' WHEN (cpuena_in='1' AND addr(23)='0') OR micro_state=io_aktion OR state="01" ELSE '0'; 
--cpuena <= '1' WHEN (cpuena_in='1' AND addr(23)='0') OR micro_state=io_aktion ELSE '0'; 
--cpuena <= '1' WHEN cpuena_in='1' AND (state(1)='0' OR addr(23)='0' OR micro_state=io_aktion) ELSE '0'; 
--cpuena <= '1' WHEN cpuena_in='1' AND (state(1)='0' OR addr(23)='0' OR IOcpuena='1') ELSE '0'; 
--cpuena <= '1' WHEN (cpuena_in='1' AND addr(23)='0') OR ROM_select='1' OR IOcpuena='1' ELSE '0'; 
--cpuena <= '1' WHEN ROM_select='1' OR USER_select <= '1' OR FLOPPY_select <= '1' OR KEY_select <= '1' ELSE
--		  IOcpuena WHEN rs232_select='1' OR ide_select='1' OR SPI_select='1' ELSE 
--		  cpuena_in; 
cpuena <= '1' WHEN ROM_select='1' OR USER_select='1' OR FLOPPY_select='1' OR PART_select='1' ELSE
		  IOcpuena WHEN rs232_select='1' OR ide_select='1' OR SPI_select='1' ELSE 
		  cpuena_in; 

rs232data <= X"FFFF" WHEN txbusy='1' ELSE X"0000";

RAM_write <= '1' when addr(23 downto 16)=X"00" AND addr(15)='0' AND state="11" ELSE '0';
ROM_select <= '1' when addr(23 downto 16)=X"00" AND addr(15)='0' ELSE '0';
rs232_select <= '1' when addr(23 downto 12)=X"DA8" ELSE '0';
--ide_select <= '1' when addr(23 downto 12)=X"DA0" ELSE '0';
ide_select <= '1' when addr(23 downto 14)="1101101000" ELSE '0';
KEY_select <= '1' when addr(23 downto 12)=X"DE0" ELSE '0';
USER_select <= '1' when addr(23 downto 12)=X"DE6" ELSE '0';
FLOPPY_select <= '1' when addr(23 downto 12)=X"DE8" ELSE '0';
PART_select <= '1' when addr(23 downto 12)=X"DEE" ELSE '0';
SPI_select <= '1' when addr(23 downto 12)=X"DA4" ELSE '0';

--idedata <= cpudata_in(7 downto 0)&cpudata_in(15 downto 8) WHEN ide_wrena='1' ELSE "ZZZZZZZZZZZZZZZZ";
idedata <= cpudata_in(7 downto 0)&cpudata_in(15 downto 8);
idea <= addr(4 downto 2);
--ide_cs0 <= '0' WHEN ide_select='1' AND addr(12)='0' ELSE '1';
--ide_cs1 <= '0' WHEN ide_select='1' AND addr(13)='0' ELSE '1';
--ide_wr <= '0' WHEN state="11" AND ide_select='1' ELSE'1';
--ide_rd <= '0' WHEN state="10" AND ide_select='1' ELSE'1';
--ide_wrena <= '1' WHEN state="11" AND ide_select='1' AND (addr(12)='0' OR addr(13)='0') ELSE '0';

--io_select <= ide_select OR rs232_select OR KEY_select;



-----------------------------------------------------------------
-- SPI-Interface
-----------------------------------------------------------------	
	sd_cs <= NOT scs;
	sd_clk <= NOT sck;
--	sd_di_in <= sd_di;
	sd_do <= sd_out(7);
	SD_busy <= shiftcnt(3);
	
	PROCESS (cpuclk, n_reset) BEGIN
		IF (cpuclk'event AND cpuclk='0') THEN
			sd_di_in <= sd_di;
		END IF;
		IF n_reset ='0' THEN 
			shiftcnt <= (OTHERS => '0');
			spi_div <= (OTHERS => '0');
--			scs(1) <= '0';
			scs <= (OTHERS => '0');
			zscs <= '0';
			sck <= '0';
--			zexe_dir <= '0';
			spi_speed <= "00000000";
		ELSIF (cpuclk'event AND cpuclk='1') THEN
			IF SD_busy='0' THEN
				scs(1) <= zscs;
--			ELSE	
--				zexe_dir <= '0';
			END IF;
--			IF SPI_select='1' AND state="11" AND SD_busy='0' AND cpuena_in='1' THEN	--SD write
			IF SPI_select='1' AND state="11" AND cpuena='1' THEN	 --SD write
				IF addr(2)='1' THEN
					scs(0) <= not cpudata_in(0);
					IF cpudata_in(7)='1' THEN
						scs(7) <= not cpudata_in(0);
					END IF;
					IF cpudata_in(6)='1' THEN
						scs(6) <= not cpudata_in(0);
					END IF;
					IF cpudata_in(5)='1' THEN
						scs(5) <= not cpudata_in(0);
					END IF;
					IF cpudata_in(4)='1' THEN
						scs(4) <= not cpudata_in(0);
					END IF;
					IF cpudata_in(3)='1' THEN
						scs(3) <= not cpudata_in(0);
					END IF;
					IF cpudata_in(2)='1' THEN
						scs(2) <= not cpudata_in(0);
					END IF;
					IF cpudata_in(1)='1' THEN
						zscs <= not cpudata_in(0);
					END IF;
				ELSE
	--				zexe_dir <= '1';
	--				SDcommand <= cpudata_in(7 downto 0);		
	--			END IF;		
	--			IF zexe_dir='1' AND SD_busy='0' THEN	 --SD write
					spi_div <= spi_speed;
					shiftcnt <= "1111";
	--				sd_out <= SDcommand;
					sd_out <= cpudata_in(7 downto 0);
					sck <= '1';
				END IF;
			ELSE
				IF spi_div="00000000" THEN
					spi_div <= spi_speed;
					IF SD_busy='1' THEN
						IF sck='0' THEN
							IF shiftcnt(2 downto 0)/="000" THEN
								sck <='1';
							END IF;
							shiftcnt <= shiftcnt-1;
							sd_out <= sd_out(6 downto 0)&sd_di_in;
						ELSE	
							sck <='0';
						END IF;
					END IF;
				ELSE
					spi_div <= spi_div-1;
				END IF;
			END IF;		
		END IF;		
	END PROCESS;



process(cpuclk, shift)
begin
   	IF cpuclk'event AND cpuclk = '1' THEN
--	IF cpuena_in='1' THEN
		micro_state <= idle;
		ld <= '0';
		IOcpuena <= '0';
		ide_wrdata <= '0';
		CASE micro_state IS
			WHEN idle => 
				ide_wr <= '1';
				ide_rd <= '1';
				ide_cs0 <= '1';
				ide_cs1 <= '1';
				ide_wrena <='0';
				IF rs232_select='1' AND state="11" THEN
--					IF txbusy='0' AND stop='0' THEN
					IF txbusy='0' THEN
						ld <= '1';
						micro_state <= io_aktion;
						IOcpuena <= '1';
					END IF;	
				ELSIF USER_select='1' AND state="11" THEN
					IF lds='0' THEN
						user <= cpudata_in(7 downto 0);
					END IF;	
--					micro_state <= io_aktion;
--					IOcpuena <= '1';
				ELSIF FLOPPY_select='1' AND state="11" THEN
					IF lds='0' THEN
						floppy(7 downto 0) <= cpudata_in(7 downto 0);
					END IF;	
					IF uds='0' THEN
						floppy(15 downto 8) <= cpudata_in(15 downto 8);
					END IF;	
--					micro_state <= io_aktion;
--					IOcpuena <= '1';
				ELSIF ide_select='1' AND (addr(12)='0' OR addr(13)='0') THEN
					micro_state <= ide1;
					IF state="11" THEN
						ide_wrdata <= '1';
					END IF;	
					IF addr(12)='0' THEN
						ide_cs0 <= '0';
					END IF;	
					IF addr(13)='0' THEN
						ide_cs1 <= '0';
					END IF;	
				ELSIF addr(23)='1' AND state(1)='1' THEN
					micro_state <= io_aktion;
					IOcpuena <= '1';
				END IF;
					
			WHEN io_aktion => 
				micro_state <= idle;
				
			WHEN ide1 =>
				micro_state <= ide2;
				IF state="11" THEN
					ide_wr <= '0';
--					ide_wrena <='1';
				END IF;	
				IF state="10" THEN
					ide_rd <= '0';
				END IF;	
			WHEN ide2 =>
				micro_state <= ide3;
				IF state="11" THEN
					ide_wrena <='1';
				END IF;	
				
			WHEN ide3 =>
				micro_state <= ide4;
			WHEN ide4 =>
				micro_state <= ide5;
			WHEN ide5 =>
				micro_state <= ide6;
			WHEN ide6 =>
				micro_state <= ide7;
				ide_wr <= '1';
				ide_rd <= '1';
				IOcpuena <= '1';
				IDErd_data <= idedata_in;
			WHEN ide7 => 
				micro_state <= idle;
				ide_wrena <='0';
				ide_cs0 <= '1';
				ide_cs1 <= '1';
				
			WHEN OTHERS => 
				micro_state <= idle;
		END CASE;
--	END IF;	
	end if;
end process; 


TxD <= not shiftout;
process(cpuclk, shift)
begin
	if shift="0000000000" then
		txbusy <= '0';
	else
		txbusy <= '1';
	end if;

	if n_reset='0' then
		shiftout <= '0';
		shift <= "0000000000"; 
	elsif cpuclk'event and cpuclk = '1' then
		if ld = '1' then
--				shiftout <= '0';
			shift <=  '1' & cpudata_in(15 downto 8) & '0';			--STOP,MSB...LSB, START
		end if;
		if clkgen/=0 then
--			if ld = '1' and stop = '0' then
--				clkgen <= "0000000001";		
--			else		
				clkgen <= clkgen-1;
--			end if;
		else	
--			clkgen <= "1110101001";--937;		--108MHz/115200
--			clkgen <= "0011101010";--234;		--27MHz/115200
--			clkgen <= "0011111000";--249-1;		--28,7MHz/115200
--			clkgen <= "0011110101";--246-1;		--28,7MHz/115200
			clkgen <= "0001111100";--249-1;		--14,3MHz/115200
--			shiftout <= not(shift(0) or not txbusy or ld);
			shiftout <= not shift(0) and txbusy;
--			stop <= shift(0);
		   	shift <=  '0' & shift(9 downto 1);
		end if;
	end if;
end process; 

process(sclk)
begin
--	if n_reset='0' then
--		partshift <= "10000000000000000000000000000000000000000"; 
--	els
	if sclk'event and sclk = '1' then
		mdbd <= mdb(4);
		mdcd <= mdb(0);
		mdcdd <= not mdcd;
		if mdcdd='1' and mdcd='1' and partshift(0)='0' then
			partshift <= mdbd&partshift(39 downto 1);
		end if;
	end if;
end process; 

end;  
